/* _NVRM_COPYRIGHT_BEGIN_
 *
 * Copyright 2013 by NVIDIA Corporation.  All rights reserved.  All
 * information contained herein is proprietary and confidential to NVIDIA
 * Corporation.  Any use, reproduction, or disclosure without the written
 * permission of NVIDIA Corporation is prohibited.
 *
 * _NVRM_COPYRIGHT_END_
 */

/*
 * This file sets up the communication between the UVM driver and RM. RM will
 * call the UVM driver providing to it the set of OPS it supports.  UVM will
 * then return by filling out the structure with the callbacks it supports.
 */

#define  __NO_VERSION__

#include "nv-misc.h"
#include "os-interface.h"
#include "nv-linux.h"

#if defined(NV_UVM_ENABLE) || defined(NV_UVM_NEXT_ENABLE)

#include "nv_uvm_interface.h"
#include "nv_gpu_ops.h"

struct UvmGpuProvider g_nvKernelProvider = {0};
struct UvmOpsUvmEvents   *g_pNvUvmEvents = NULL;

//
// TODO: (Bug 1349097) move these UVM_* definitions out to UVM. UVM should do
// its own translation of RM error codes. It makes no sense to make RM translate
// error codes, because it shouldn't have to know the error code system of
// whatever system is calling it.
//
typedef unsigned UVM_STATUS;

#define UVM_OK                          0x00000000
#define UVM_MORE_DATA_AVAILABLE         0x40000001

// This numeric range is reserved for using Unix errno's directly:
#define UVM_RESERVED_UNIX_FIRST_ERRNO   0x00000001
#define UVM_RESERVED_UNIX_LAST_ERRNO    0x00000400

#define UVM_ERR_FAILED                  0x80000001
#define UVM_ERR_INTERNAL_ERROR          0x80000002
#define UVM_ERR_NOT_PERMITTED           0x80000003
#define UVM_ERR_INVALID_ARGUMENT        0x80000004
#define UVM_ERR_OUT_OF_MEMORY           0x80000005
#define UVM_ERR_INSUFFICIENT_RESOURCES  0x80000006
#define UVM_ERR_FAILED_DRIVER_LOAD      0x80000007
#define UVM_ERR_OVERLAPPING_COMMIT      0x80000008
#define UVM_ERR_ADDRESS_IN_USE          0x80000009
#define UVM_ERR_UNSUPPORTED_GPU         0x8000000a
#define UVM_ERR_RETRY_LATER             0x8000000b
#define UVM_ERR_NOT_SUPPORTED           0x8000000c
#define UVM_ERR_NOT_YET_IMPLEMENTED     0x8000000d
#define UVM_ERR_GPU_DMA_NOT_INITIALIZED 0x8000000e
#define UVM_ERR_INVALID_INDEX           0x8000000f


static UVM_STATUS convertToUvmErrorCode(RM_STATUS rmStatus);

static UVM_STATUS convertToUvmErrorCode(RM_STATUS rmStatus)
{
    switch(rmStatus)
    {
        case RM_OK:
            return UVM_OK;

        case RM_ERROR:
            return UVM_ERR_FAILED;

        case RM_ERR_INSUFFICIENT_RESOURCES:
            return UVM_ERR_INSUFFICIENT_RESOURCES;

        case RM_ERR_INVALID_ARGUMENT:
            return UVM_ERR_INVALID_ARGUMENT;

        case RM_ERR_NOT_SUPPORTED:
            return UVM_ERR_NOT_SUPPORTED;

        case RM_ERR_INVALID_INDEX:
            return UVM_ERR_INVALID_INDEX;

        default:
            return UVM_ERR_FAILED;
    }
}

UVM_STATUS nvUvmInterfaceSessionCreate(uvmGpuSessionHandle *session)
{
    nv_stack_t *sp = NULL;
    RM_STATUS status;

    NV_KMEM_CACHE_ALLOC_STACK(sp);
    if (sp == NULL)
        return UVM_ERR_INVALID_ARGUMENT;

    status = rm_gpu_ops_create_session(sp, (gpuSessionHandle *)session);

    NV_KMEM_CACHE_FREE_STACK(sp);
    return convertToUvmErrorCode(status);
}


UVM_STATUS nvUvmInterfaceSessionDestroy(uvmGpuSessionHandle session)
{
    nv_stack_t *sp = NULL;
    RM_STATUS status;

    NV_KMEM_CACHE_ALLOC_STACK(sp);
    if (sp == NULL)
        return RM_ERROR;

    status = rm_gpu_ops_destroy_session(sp, (gpuSessionHandle)session);

    NV_KMEM_CACHE_FREE_STACK(sp);
    return convertToUvmErrorCode(status);
}

UVM_STATUS nvUvmInterfaceAddressSpaceCreateMirrored(uvmGpuSessionHandle session,
                                             unsigned long long uuidMsb,
                                             unsigned long long uuidLsb,
                                             uvmGpuAddressSpaceHandle *vaSpace)
{
    nv_stack_t *sp = NULL;
    RM_STATUS status;

    NV_KMEM_CACHE_ALLOC_STACK(sp);
    if (sp == NULL)
        return RM_ERROR;

    status = rm_gpu_ops_address_space_create_mirrored(
             sp, (gpuSessionHandle)session,
             uuidMsb, uuidLsb, (gpuAddressSpaceHandle *)vaSpace);

    NV_KMEM_CACHE_FREE_STACK(sp);
    return convertToUvmErrorCode(status);
}

UVM_STATUS nvUvmInterfaceAddressSpaceCreate(uvmGpuSessionHandle session,
                                            unsigned long long uuidMsb,
                                            unsigned long long uuidLsb,
                                            uvmGpuAddressSpaceHandle *vaSpace)
{
    nv_stack_t *sp = NULL;
    RM_STATUS status;

    NV_KMEM_CACHE_ALLOC_STACK(sp);
    if (sp == NULL)
        return RM_ERROR;

    status = rm_gpu_ops_address_space_create(
             sp, (gpuSessionHandle)session,
             uuidMsb, uuidLsb, (gpuAddressSpaceHandle *)vaSpace);

    NV_KMEM_CACHE_FREE_STACK(sp);
    return convertToUvmErrorCode(status);
}

void nvUvmInterfaceAddressSpaceDestroy(uvmGpuAddressSpaceHandle vaSpace)
{
    nv_stack_t *sp = NULL;

    NV_KMEM_CACHE_ALLOC_STACK(sp);
    if (sp == NULL)
        return;

    rm_gpu_ops_address_space_destroy(
        sp, (gpuAddressSpaceHandle)vaSpace);

    NV_KMEM_CACHE_FREE_STACK(sp);
    return;
}

UVM_STATUS nvUvmInterfaceMemoryAllocFB(uvmGpuAddressSpaceHandle vaSpace,
                    NvLength length, UvmGpuPointer * gpuPointer)
{
    nv_stack_t *sp = NULL;
    RM_STATUS status;

    NV_KMEM_CACHE_ALLOC_STACK(sp);
    if (sp == NULL)
        return RM_ERROR;

    status = rm_gpu_ops_memory_alloc_fb(
             sp, (gpuAddressSpaceHandle)vaSpace,
             (NvLength)length, (NvU64 *) gpuPointer);

    NV_KMEM_CACHE_FREE_STACK(sp);
    return convertToUvmErrorCode(status);
}

UVM_STATUS nvUvmInterfaceMemoryAllocSys(uvmGpuAddressSpaceHandle vaSpace,
                    NvLength length, UvmGpuPointer * gpuPointer)
{
    nv_stack_t *sp = NULL;
    RM_STATUS status;

    NV_KMEM_CACHE_ALLOC_STACK(sp);
    if (sp == NULL)
        return RM_ERROR;

    status = rm_gpu_ops_memory_alloc_sys(
             sp, (gpuAddressSpaceHandle)vaSpace,
             (NvLength)length, (NvU64 *) gpuPointer);

    NV_KMEM_CACHE_FREE_STACK(sp);
    return convertToUvmErrorCode(status);
}

void nvUvmInterfaceMemoryFree(uvmGpuAddressSpaceHandle vaSpace,
                    UvmGpuPointer gpuPointer)
{
    nv_stack_t *sp = NULL;
    RM_STATUS status;

    NV_KMEM_CACHE_ALLOC_STACK(sp);
    if (sp == NULL)
        return;

    status = rm_gpu_ops_memory_free(
             sp, (gpuAddressSpaceHandle)vaSpace,
             (NvU64) gpuPointer);

    NV_KMEM_CACHE_FREE_STACK(sp);
    return;
}

UVM_STATUS nvUvmInterfaceMemoryCpuMap(uvmGpuAddressSpaceHandle vaSpace,
           UvmGpuPointer gpuPointer, NvLength length, void **cpuPtr)
{
    nv_stack_t *sp = NULL;
    RM_STATUS status;

    NV_KMEM_CACHE_ALLOC_STACK(sp);
    if (sp == NULL)
        return RM_ERROR;

    status = rm_gpu_ops_memory_cpu_map(
             sp, (gpuAddressSpaceHandle)vaSpace,
             (NvU64) gpuPointer, (NvLength) length, cpuPtr);

    NV_KMEM_CACHE_FREE_STACK(sp);
    return convertToUvmErrorCode(status);
}

void nvUvmInterfaceMemoryCpuUnMap(uvmGpuAddressSpaceHandle vaSpace,
           void *cpuPtr)
{
    nv_stack_t *sp = NULL;
    RM_STATUS status;

    NV_KMEM_CACHE_ALLOC_STACK(sp);
    if (sp == NULL)
        return;

    status = rm_gpu_ops_memory_cpu_ummap(
             sp, (gpuAddressSpaceHandle)vaSpace,
             cpuPtr);

    NV_KMEM_CACHE_FREE_STACK(sp);
    return;
}

UVM_STATUS nvUvmInterfaceChannelAllocate(uvmGpuAddressSpaceHandle  vaSpace,
                     uvmGpuChannelHandle *channel,
                     UvmGpuChannelPointers * pointers)
{
    nv_stack_t *sp = NULL;
    RM_STATUS status;
    struct gpuChannelInfo *channelInfo = NULL;

    NV_KMEM_CACHE_ALLOC_STACK(sp);
    if (sp == NULL)
        return RM_ERROR;

    status = os_alloc_mem((void **)&channelInfo, sizeof(*channelInfo));
    if (status != RM_OK)
        goto cleanup_stack;

    status = rm_gpu_ops_channel_allocate(
             sp, (gpuAddressSpaceHandle)vaSpace, (gpuChannelHandle *)channel,
             channelInfo);

    if(status != RM_OK)
        goto cleanup_chInfo;

    pointers->GPGet = channelInfo->GPGet;
    pointers->GPPut = channelInfo->GPPut;
    pointers->gpFifoEntries = channelInfo->gpFifoEntries;
    pointers->channelClassNum = channelInfo->channelClassNum;
    pointers->numGpFifoEntries = channelInfo->numGpFifoEntries;
    pointers->errorNotifier = channelInfo->errorNotifier;

 cleanup_chInfo:
    os_free_mem(channelInfo);
 cleanup_stack:
    NV_KMEM_CACHE_FREE_STACK(sp);
    return convertToUvmErrorCode(status);
}

void nvUvmInterfaceChannelDestroy(uvmGpuChannelHandle channel)
{
    nv_stack_t *sp = NULL;
    RM_STATUS status;

    NV_KMEM_CACHE_ALLOC_STACK(sp);
    if (sp == NULL)
        return;

    status = rm_gpu_ops_channel_destroy(
                    sp, (gpuChannelHandle)channel);
    if(status != RM_OK)
        goto cleanup;

cleanup:
    NV_KMEM_CACHE_FREE_STACK(sp);
    return;

}

const char* nvUvmInterfaceChannelTranslateError(unsigned info32)
{
    nv_stack_t *sp = NULL;
    const char *errorString;

    NV_KMEM_CACHE_ALLOC_STACK(sp);
    if (sp == NULL)
        return 0;

    errorString = rm_gpu_ops_channel_translate_error(
                  sp, info32);

    NV_KMEM_CACHE_FREE_STACK(sp);
    return errorString;
}

UVM_STATUS nvUvmInterfaceCopyEngineAllocate(
                     uvmGpuChannelHandle channel,
                     unsigned indexStartingAtOne,
                     unsigned * copyEngineClassNumber,
                     uvmGpuCopyEngineHandle *copyEngine)
{
    nv_stack_t *sp = NULL;
    RM_STATUS status;

    NV_KMEM_CACHE_ALLOC_STACK(sp);
    if (sp == NULL)
        return RM_ERROR;

    status = rm_gpu_ops_copy_engine_allocate(
             sp, (gpuChannelHandle) channel, indexStartingAtOne,
             copyEngineClassNumber, (gpuObjectHandle *) copyEngine);


    NV_KMEM_CACHE_FREE_STACK(sp);
    return convertToUvmErrorCode(status);
}

UVM_STATUS nvUvmInterfaceGetAttachedUuids(NvU8 *pUuidList, unsigned *numGpus)
{
    nv_stack_t *sp = NULL;
    RM_STATUS status;

    NV_KMEM_CACHE_ALLOC_STACK(sp);
    if (sp == NULL)
        return RM_ERROR;

    status = rm_gpu_ops_get_attached_uuids(sp, pUuidList, numGpus);

    NV_KMEM_CACHE_FREE_STACK(sp);
    return convertToUvmErrorCode(status);
}

// this function is called by the UVM driver to register the ops
UVM_STATUS nvUvmInterfaceRegisterUvmOps(struct UvmGpuProvider **ppGpuProvider,
                                        struct UvmOpsUvmEvents *importedUvmOps)
{
    if (!ppGpuProvider || !importedUvmOps)
    {
        return UVM_ERR_INVALID_ARGUMENT;
    }

    g_nvKernelProvider.sessionCreate = nvUvmInterfaceSessionCreate;
    g_nvKernelProvider.sessionDestroy = nvUvmInterfaceSessionDestroy;
    g_nvKernelProvider.addressSpaceCreate = nvUvmInterfaceAddressSpaceCreate;
    g_nvKernelProvider.addressSpaceCreateMirrored = nvUvmInterfaceAddressSpaceCreateMirrored;
    g_nvKernelProvider.addressSpaceDestroy = nvUvmInterfaceAddressSpaceDestroy;
    g_nvKernelProvider.allocGpuMemoryFB = nvUvmInterfaceMemoryAllocFB;
    g_nvKernelProvider.allocGpuMemorySys = nvUvmInterfaceMemoryAllocSys;
    g_nvKernelProvider.freeGpuMemory = nvUvmInterfaceMemoryFree;
    g_nvKernelProvider.cpuMap = nvUvmInterfaceMemoryCpuMap;
    g_nvKernelProvider.cpuUnmap = nvUvmInterfaceMemoryCpuUnMap;
    g_nvKernelProvider.channelAllocate = nvUvmInterfaceChannelAllocate;
    g_nvKernelProvider.channelDestroy  = nvUvmInterfaceChannelDestroy;
    g_nvKernelProvider.channelTranslateError = nvUvmInterfaceChannelTranslateError;
    g_nvKernelProvider.copyEngineAllocate = nvUvmInterfaceCopyEngineAllocate;
    g_nvKernelProvider.getAttachedUuids = nvUvmInterfaceGetAttachedUuids;

    *ppGpuProvider = &g_nvKernelProvider;
    g_pNvUvmEvents = importedUvmOps;

    return UVM_OK;
}
EXPORT_SYMBOL(nvUvmInterfaceRegisterUvmOps);

UVM_STATUS nvUvmInterfaceDeRegisterUvmOps(void)
{
    memset(&g_nvKernelProvider, 0, sizeof(struct UvmGpuProvider));
    g_pNvUvmEvents = NULL;

    return UVM_OK;
}
EXPORT_SYMBOL(nvUvmInterfaceDeRegisterUvmOps);


void nv_uvm_notify_start_device(NvU8 *pUuid)
{
    UvmGpuUuid uvmUuid;

    memcpy(uvmUuid.uuid, pUuid, UVM_UUID_LEN);

    if(g_pNvUvmEvents && g_pNvUvmEvents->startDevice)
    {
        g_pNvUvmEvents->startDevice(&uvmUuid);
    }
    return;
}

void nv_uvm_notify_stop_device(NvU8 *pUuid)
{
    UvmGpuUuid uvmUuid;

    memcpy(uvmUuid.uuid, pUuid, UVM_UUID_LEN);

    if(g_pNvUvmEvents && g_pNvUvmEvents->stopDevice)
    {
        g_pNvUvmEvents->stopDevice(&uvmUuid);
    }
    return;
}

RM_STATUS nv_uvm_event_interrupt(void)
{
    //
    // Since there is no API to de-register/modify the callback,
    // we can safely compare and invoke if it is non-null.
    //
    if (g_pNvUvmEvents && g_pNvUvmEvents->isrTopHalf)
        return g_pNvUvmEvents->isrTopHalf();

    //
    // RM_OK means that the interrupt was for the UVM driver, so use
    // RM_ERR_NO_INTR_PENDING to tell the caller that we didn't do anything.
    //
    return RM_ERR_NO_INTR_PENDING;
}

#endif // NV_UVM_ENABLE || NV_UVM_NEXT_ENABLE
