/*******************************************************************************
    Copyright (c) 2013 NVidia Corporation

    Permission is hereby granted, free of charge, to any person obtaining a copy
    of this software and associated documentation files (the "Software"), to
    deal in the Software without restriction, including without limitation the
    rights to use, copy, modify, merge, publish, distribute, sublicense, and/or
    sell copies of the Software, and to permit persons to whom the Software is
    furnished to do so, subject to the following conditions:

        The above copyright notice and this permission notice shall be
        included in all copies or substantial portions of the Software.

    THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
    IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
    FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL
    THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
    LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
    FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
    DEALINGS IN THE SOFTWARE.

*******************************************************************************/

/*
 * nvidia_uvm_common.c
 *
 * This file contains code that is common to all variants of the (Linux) UVM
 * kernel module.
 *
 * Initially, UVM-Lite and Uvm-Next work is proceeding somewhat
 * independently, but the long- and short-term plan is for them to share the
 * same kernel module (this one). That's why, at least initially, you'll find
 * files for UVM-Lite, for Uvm-Next, and for common code. We'll refactor as
 * necessary, to achieve the right balance of sharing code and decoupling the
 * features.
 *
 * --jhubbard
 *
 */

#include "uvm_gpu_provider.h"
#include "nvidia_uvm_common.h"
#include "nvidia_uvm_utils.h"
#include "nvidia_uvm_linux.h"

#if defined(NVIDIA_UVM_LITE_ENABLED)
#include "nvidia_uvm_lite.h"
#else
static int uvmlite_init(dev_t uvmBaseDev)
{
    return 0;
}
static void uvmlite_exit(void)
{

}
static RM_STATUS uvmlite_gpu_event_start_device(UvmGpuUuid *gpuUuidStruct)
{
    return RM_OK;
}
static RM_STATUS uvmlite_gpu_event_stop_device(UvmGpuUuid *gpuUuidStruct)
{
    return RM_OK;
}
#endif // NVIDIA_UVM_LITE_ENABLED

#if defined(NVIDIA_UVM_NEXT_ENABLED)
#include "nvidia_uvm_next.h"
#else
static int uvmnext_init(dev_t uvmBaseDev)
{
    return 0;
}
static void uvmnext_exit(void)
{

}
static RM_STATUS uvmnext_isr_top_half(void)
{
    return RM_ERR_NO_INTR_PENDING;
}
static RM_STATUS uvmnext_gpu_event_start_device(UvmGpuUuid *gpuUuidStruct)
{
    return RM_OK;
}
static RM_STATUS uvmnext_gpu_event_stop_device(UvmGpuUuid *gpuUuidStruct)
{
    return RM_OK;
}
#endif // NVIDIA_UVM_NEXT_ENABLED

static dev_t g_uvmBaseDev;

// This is a single GPU provider. We need to enhance to a list
static struct kmem_cache * g_uvmGpuProviderTrackingCache __read_mostly = NULL;
static UvmGpuProviderTracking *g_pProviderList = NULL;

static UvmGpuProviderTracking* allocGpuProvider(void)
{
    UvmGpuProviderTracking *pProvider = NULL;
    pProvider = (UvmGpuProviderTracking*)kmem_cache_alloc(
                                         g_uvmGpuProviderTrackingCache,
                                         GFP_KERNEL);
    if (pProvider)
    {
        memset(pProvider, 0 , sizeof(*pProvider));
    }

    return pProvider;
}

// TODO: This would be easier if RM allowed for multiple registrations, since we
//       could register UVM-Lite and UVM-Next separately (bug 1372835).
static RM_STATUS uvm_gpu_event_start_device(UvmGpuUuid *gpuUuidStruct)
{
    RM_STATUS rmStatus;

    if (uvmlite_enabled())
    {
        rmStatus = uvmlite_gpu_event_start_device(gpuUuidStruct);
        if (rmStatus != RM_OK)
            return rmStatus;
    }

    if (uvmnext_enabled())
    {
        rmStatus = uvmnext_gpu_event_start_device(gpuUuidStruct);
        if (rmStatus != RM_OK)
        {
            // Take UVM-Lite back down if UVM-Next failed to start
            if (uvmlite_enabled())
                uvmlite_gpu_event_stop_device(gpuUuidStruct);
            return rmStatus;
        }
    }

    return RM_OK;
}

static RM_STATUS uvm_gpu_event_stop_device(UvmGpuUuid *gpuUuidStruct)
{
    RM_STATUS uvmliteStatus = RM_OK, uvmnextStatus = RM_OK;

    // Stop both of them even when one returns an error
    if (uvmlite_enabled())
        uvmliteStatus = uvmlite_gpu_event_stop_device(gpuUuidStruct);

    if (uvmnext_enabled())
        uvmnextStatus = uvmnext_gpu_event_stop_device(gpuUuidStruct);

    if (uvmliteStatus != RM_OK)
        return uvmliteStatus;

    return uvmnextStatus;
}

static UVM_STATUS uvmSetupGpuProvider(UvmGpuProviderTracking *pProvider)
{
    UVM_STATUS status = UVM_OK;

#ifdef NVIDIA_UVM_RM_ENABLED
    pProvider->exportedUvmOps.startDevice = uvm_gpu_event_start_device;
    pProvider->exportedUvmOps.stopDevice  = uvm_gpu_event_stop_device;
    pProvider->exportedUvmOps.isrTopHalf  = uvmnext_isr_top_half;

    // call RM to exchange the function pointers.
    status = nvUvmInterfaceRegisterUvmOps(&(pProvider->registeredGpuOps),
                                          &pProvider->exportedUvmOps);
#endif
    return status;
}

static int __init uvm_init(void)
{
    int ret = 0;
    int devAlloced = 0;
    int uvmliteInitialized = 0;
    int uvmnextInitialized = 0;
    int registered = 0;
    UVM_STATUS uvmStatus;

    // The various helper init routines will create their own minor devices, so
    // we only need to create space for them here.
    ret = alloc_chrdev_region(&g_uvmBaseDev,
                              0,
                              NVIDIA_UVM_NUM_MINOR_DEVICES,
                              NVIDIA_UVM_DEVICE_NAME);
    if (ret)
    {
        pr_err("%s: alloc_chrdev_region failed: %d\n", __FUNCTION__, ret);
        goto error;
    }
    devAlloced = 1;

    if (uvmlite_enabled())
    {
        ret = uvmlite_init(g_uvmBaseDev);
        if (ret)
            goto error;
        uvmliteInitialized = 1;
    }

    if (uvmnext_enabled())
    {
        ret = uvmnext_init(g_uvmBaseDev);
        if (ret)
            goto error;
        uvmnextInitialized = 1;
    }

    NV_KMEM_CACHE_CREATE(g_uvmGpuProviderTrackingCache,
                         "uvm_provider_tracking_t",
                         struct UvmGpuProviderTracking_tag);
    if (!g_uvmGpuProviderTrackingCache)
    {
        pr_err("%s: Provider tracking alloc failed\n", __FUNCTION__);
        ret = -ENOMEM;
        goto error;
    }
    
    // Register with RM
    //
    // TODO: This should probably happen before calling uvmlite_init and
    //       uvmnext_init, so those routines can make RM calls as part of their
    //       setup. But doing that means we could get event callbacks before
    //       those init routines are called, which is scary. Ideally this could
    //       be solved by having RM handle multiple registrations (bug 1372835).
    if (rm_enabled())
    {
        g_pProviderList = allocGpuProvider();
        if (!g_pProviderList)
            goto error;

        uvmStatus = uvmSetupGpuProvider(getProvider());
        if (uvmStatus != UVM_OK)
        {
            pr_err("Cannot register ops with NVIDIA driver. UVM error: 0x%x\n",
                   uvmStatus);
            // TODO: Need a helper function to convert from UVM_STATUS to errno
            ret = -EINVAL;
            goto error;
        }
        registered = 1;
    }

    // Register with RM
    //
    // TODO: Ideally this should be called within uvmlite_init but this is related to 
    //       (bug 1372835) where we dont want to get event callbacks before init.

    if(uvmliteInitialized && registered)
    {
        ret = uvmlite_setup_gpu_list();
        if (ret)
        {
            pr_err("%s: uvm_lite_setup_gpu_list failed\n", __FUNCTION__);
            goto error; 
        }
    }

    pr_info("Loaded the UVM driver, major device number %d\n",
            MAJOR(g_uvmBaseDev));

    return 0;

error:
#ifdef NVIDIA_UVM_RM_ENABLED
    if (registered)
        nvUvmInterfaceDeRegisterUvmOps();
#endif

    if (g_pProviderList)
        kmem_cache_free(g_uvmGpuProviderTrackingCache, g_pProviderList);
    kmem_cache_destroy_safe(g_uvmGpuProviderTrackingCache);

    if (uvmliteInitialized)
        uvmlite_exit();

    if (uvmnextInitialized)
        uvmnext_exit();

    if (devAlloced)
        unregister_chrdev_region(g_uvmBaseDev, NVIDIA_UVM_NUM_MINOR_DEVICES);

    return ret;
}

static void __exit uvm_exit(void)
{
#ifdef NVIDIA_UVM_RM_ENABLED
    // this will cleanup registered interfaces
    nvUvmInterfaceDeRegisterUvmOps();
#endif

    if (g_pProviderList)
        kmem_cache_free(g_uvmGpuProviderTrackingCache, g_pProviderList);
    kmem_cache_destroy_safe(g_uvmGpuProviderTrackingCache);

    if (uvmlite_enabled())
        uvmlite_exit();

    if (uvmnext_enabled())
        uvmnext_exit();

    unregister_chrdev_region(g_uvmBaseDev, NVIDIA_UVM_NUM_MINOR_DEVICES);

    pr_info("Unregistered the UVM driver\n");
}

UvmGpuProviderTracking* getProvider(void)
{
    return g_pProviderList;
}

void uvm_print_uuid(const char * function, const char * msg,
                    UvmGpuUuid * pGpuUuidStruct)
{
    char uuidBuffer[UVM_GPU_UUID_TEXT_BUFFER_LENGTH];

    format_uuid_to_buffer(uuidBuffer, sizeof(uuidBuffer), pGpuUuidStruct);

    pr_devel("%s: %s: %s\n", function, msg, uuidBuffer);
}

module_init(uvm_init);
module_exit(uvm_exit);
MODULE_LICENSE("MIT");
MODULE_INFO(supported, "external");
