/*******************************************************************************
    Copyright (c) 2013 NVidia Corporation

    Permission is hereby granted, free of charge, to any person obtaining a copy
    of this software and associated documentation files (the "Software"), to
    deal in the Software without restriction, including without limitation the
    rights to use, copy, modify, merge, publish, distribute, sublicense, and/or
    sell copies of the Software, and to permit persons to whom the Software is
    furnished to do so, subject to the following conditions:

        The above copyright notice and this permission notice shall be
        included in all copies or substantial portions of the Software.

    THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
    IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
    FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL
    THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
    LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
    FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
    DEALINGS IN THE SOFTWARE.

*******************************************************************************/

//
// nvidia_uvm_lite.h
//
// This file contains declarations for UVM-Lite code.
//
//

#ifndef _NVIDIA_UVM_LITE_H
#define _NVIDIA_UVM_LITE_H

#include "uvmtypes.h"
#include "uvm_ioctl.h"
#include "nvidia_uvm_linux.h"
#include "rmretval.h"

#define NVIDIA_UVM_LITE_MINOR_NUMBER 0

#define UVM_INVALID_HOME_GPU_INDEX     0xDEADBEEF
#define UVM_INVALID_PER_PROCESS_PER_GPU_COUNTER_INDEX 0xFFFFFFFF

// Forward declarations:
struct UvmCommitRecord_tag;
struct UvmPerProcessGpuMigs_tag;
struct UvmGpuMigrationTracking_tag;

#define UVM_MAX_STREAMS 256
#define UVM_STREAMS_CACHE_SIZE 1024
#define UVM_MAX_SESSIONS_PER_PROCESS 64


static const uid_t UVM_ROOT_UID = 0;
//
// UVM Counter memory descriptors and users information
//
typedef struct UvmCounterInfo_tag
{
    // physical page *
    struct page *pCounterPage;

    // kernel mapping of above page
    unsigned long long *sysAddr;

    // Number of enabled sessions for each counter
    atomic_t sessionCount[UVM_TOTAL_COUNTERS];

} UvmCounterInfo;

typedef struct UvmProcessCounterInfo_tag
{
    UvmCounterInfo procSingleGpuCounterArray[UVM_MAX_GPUS];
    UvmCounterInfo procAllGpuCounter;
    atomic_t refcountUsers;
    //
    // There can only be UVM_MAX_GPUS number of GPUs in a system, and each GPU
    // has a unique UUID. Therefore, the design here dictates that each process
    // can add new GPUs to its gpuHistory[] array, but no duplicates are
    // allowed, and entries are never deleted from the array.
    //
    UvmGpuUuid gpuHistory[UVM_MAX_GPUS];
    // number of currently assigned gpuHistory entries:
    unsigned numGpus;
    // In order to add new entries, callers must acquire this lock:
    struct rw_semaphore  gpuHistoryLock;
} UvmProcessCounterInfo;

//
// This structure saves per session information.
//
typedef struct UvmSessionInfo_tag
{
    unsigned pidSessionOwner;
    unsigned pidTarget;
    //
    // effective user id of referenced processRecord
    // stored for checking privilages
    //
    uid_t euidTarget;
    //
    // user base address for mapping counter pages.
    // First page length address contains per process all gpu counters,
    // others (per process per gpu counters page) are shifted by 1.
    //
    unsigned long long *mappedUserBaseAddress;
    UvmProcessCounterInfo *pTargetCounterInfo;
} UvmSessionInfo;

//
// UvmPerProcessGpuMigs struct: there is a one-to-one relationship between
// processes that call into the UVM-Lite kernel driver, and this data structure.
//
//
typedef struct UvmPerProcessGpuMigs_tag
{
    //
    // This GPU UUID is here in order to aid reading counters. It is actually
    // needed for operation, even though we use
    // UvmCommitRecord.cachedHomeGpuPerProcessIndex e.g. for quick counter
    // increment.
    //
    UvmGpuUuid  migUuid;
    struct UvmGpuMigrationTracking_tag * migTracker;
}UvmPerProcessGpuMigs;

typedef struct UvmStreamRecord_tag UvmStreamRecord;

typedef struct UvmProcessRecord_tag
{
    //
    //  TODO:
    //      1. "Non-idle" stream list should be stored here.
    //          This in turns allows all "Regions" associated with this process
    //          to be enumerated. Probably going to be useful for the (shared)
    //          bits of the debug/profiler infrastructure
    //      2. Complete list of all GpuRecords for this process
    //
    UvmPerProcessGpuMigs gpuMigs[UVM_MAX_GPUS];
    UvmProcessCounterInfo *pProcessCounterInfo;
    // effective user id of this process, for security check
    uid_t euid;

    // List of all streams TODO: should probably be moved to a hashmap
    struct list_head allStreamList;
    // Trivial cache storing recently used streams at streamId mod cache size
    UvmStreamRecord *streamsCache[UVM_STREAMS_CACHE_SIZE];

    //
    // Number of streams (other than NO and ALL streams) in the running state
    // Used for tracking when to start/stop the ALL stream
    //
    UvmSize runningStreams;
    // Per Process Session array.
    UvmSessionInfo sessionInfoArray[UVM_MAX_SESSIONS_PER_PROCESS];
    // This lock is used to protect sessionInfoArray
    struct rw_semaphore sessionInfoLock;

    unsigned pid;
    // Entry in g_uvmProcessRecordList
    struct list_head processRecordNode;
}UvmProcessRecord;

struct UvmStreamRecord_tag
{
    UvmProcessRecord *processRecord;
    UvmStream streamId;
    UvmBool isRunning;
    struct list_head allStreamListNode;
    struct list_head commitRecordsList;
};

//
//  Tracks committed regions of memory
//
typedef struct UvmCommitRecord_tag
{
    //
    // Invariant: The home GPU might stop running, or get an RC recovery that
    // kills the channel that this commit record uses. However, the home GPU's
    // identity will not change.
    //
    // As it turns out, all we actually need is the UUID of this home GPU.
    //
    UvmGpuUuid homeGpuUuid; // GPU that owns the allocation

    //
    // This is used to index into UvmProcessRecord.gpuMigs[].  Allowed values
    // are:
    //
    //     1. Zero through (UVM_MAX_GPUS-1)
    //
    //     2. UVM_INVALID_HOME_GPU_INDEX
    //
    // If it is UVM_INVALID_HOME_GPU_INDEX, then it must not be used.
    //
    unsigned         cachedHomeGpuPerProcessIndex;
    unsigned         cachedPerProcessPerGpuCounterIndex;

    UvmUPtr          baseAddress;   // immutable, must be page aligned
    UvmSize          length;        // immutable, must be page aligned

    struct DriverPrivate_tag * osPrivate;
    struct vm_area_struct * vma;
    struct UvmPageTracking_tag ** commitRecordPages;

    UvmStreamRecord * pStream;
    struct list_head streamRegionsListNode;
}UvmCommitRecord;

typedef struct DriverPrivate_tag
{
    struct list_head       pageList;
    UvmProcessRecord  processRecord;
    struct rw_semaphore    uvmPrivLock;
    struct file            *privFile;
}DriverPrivate;

struct file;

//
//
// UVM-Lite char driver entry points:
//
int uvmlite_init(dev_t uvmBaseDev);
void uvmlite_exit(void);
int uvmlite_setup_gpu_list(void);

UVM_STATUS uvmlite_set_stream_running(DriverPrivate* pPriv, UvmStream streamId);
UVM_STATUS uvmlite_set_streams_stopped(DriverPrivate* pPriv,
                                      UvmStream * streamIdArray,
                                      UvmSize nStreams);
UVM_STATUS uvmlite_region_set_stream(UvmCommitRecord *pRecord,
    UvmStream streamId);
UVM_STATUS uvmlite_migrate_to_gpu(unsigned long long baseAddress,
                                  UvmSize length,
                                  unsigned migrateFlags,
                                  struct vm_area_struct *vma,
                                  UvmCommitRecord * pRecord);
//
// UVM-Lite core functionality:
//
UvmCommitRecord *
uvmlite_create_commit_record(UVM_REGION_COMMIT_PARAMS *pParams,
                             DriverPrivate *pPriv,
                             struct vm_area_struct *vma);

//
// UVM-Lite page cache:
//
typedef struct UvmPageTracking_tag
{
    struct page *uvmPage;
    struct list_head pageListNode;
}UvmPageTracking;

int uvm_page_cache_init(void);
void uvm_page_cache_destroy(void);

UvmPageTracking * uvm_page_cache_alloc_page(DriverPrivate* pPriv);
void uvm_page_cache_free_page(UvmPageTracking *pTracking,
                              DriverPrivate *pPriv,
                              const char *caller);

void uvm_page_cache_free_for_this_process(DriverPrivate* pPriv,
                                          const char * caller);
//
// UVM-Lite outer layer: UVM API calls that are issued via ioctl() call:
//
UVM_STATUS uvmAPI_ReserveVa(UVM_RESERVE_VA_PARAMS *pParams, struct file *filp);
UVM_STATUS uvmAPI_ReleaseVa(UVM_RELEASE_VA_PARAMS *pParams, struct file *filp);
UVM_STATUS uvmAPI_RegionCommit(UVM_REGION_COMMIT_PARAMS *pParams,
                               struct file *filp);
UVM_STATUS uvmAPI_RegionDecommit(UVM_REGION_DECOMMIT_PARAMS *pParams,
                                 struct file *filp);
UVM_STATUS uvmAPI_RegionSetStream(UVM_REGION_SET_STREAM_PARAMS *pParams,
                                  struct file *filp);
UVM_STATUS uvmAPI_SetStreamRunning(UVM_SET_STREAM_RUNNING_PARAMS *pParams,
                                   struct file *filp);
UVM_STATUS uvmAPI_SetStreamStopped(UVM_SET_STREAM_STOPPED_PARAMS *pParams,
                                   struct file *filp);
UVM_STATUS uvmAPI_MigrateToGpu(UVM_MIGRATE_TO_GPU_PARAMS *pParams,
                               struct file *filp);
#if defined (DEBUG)
UVM_STATUS uvmAPI_RunTest(UVM_RUN_TEST_PARAMS *pParams, struct file *filp);
#endif
UVM_STATUS uvmAPI_AddSession(UVM_ADD_SESSION_PARAMS *pParams,
                             struct file *filp);
UVM_STATUS uvmAPI_RemoveSession(UVM_REMOVE_SESSION_PARAMS *pParams,
                                struct file *filp);
UVM_STATUS uvmAPI_EnableCounters(UVM_ENABLE_COUNTERS_PARAMS *pParams,
                                 struct file *filp);
UVM_STATUS uvmAPI_MapCounter(UVM_MAP_COUNTER_PARAMS *pParams,
                             struct file *filp);
//
// UVM-Lite counters routines
//
UvmProcessCounterInfo *
uvmlite_get_and_refcount_process_counter_info(UvmProcessRecord *pProcessRecord);

void uvmlite_put_and_unrefcount_process_counter_info
(
    UvmProcessCounterInfo *pProcessCounterInfo
);

UVM_STATUS UvmGetGpuIndex(UvmProcessCounterInfo *pProcessCounterInfo,
                          UvmGpuUuid *gpuUuid, int *gpuIndex);

UVM_STATUS uvmlite_init_process_counters(UvmProcessRecord *pRecord);

void uvmlite_free_process_counters(UvmProcessCounterInfo *pProcCounterInfo);

UVM_STATUS uvmlite_secure_get_process_counter_info(unsigned pidTarget,
     UvmProcessCounterInfo **ppProcessCounterInfo, uid_t *pEuid);

RM_STATUS uvmlite_gpu_event_start_device(UvmGpuUuid *gpuUuidStruct);
RM_STATUS uvmlite_gpu_event_stop_device(UvmGpuUuid *gpuUuidStruct);

#endif // _NVIDIA_UVM_LITE_H

